@@ -1,10 +1,10 @@
-import sys
-import os
 import re
 from datetime import datetime
 import hashlib
 import time
 import stat
+from urllib.parse import urlencode
+from io import BytesIO
 
 import pycurl
 import ftputil
@@ -12,60 +12,10 @@ import ftputil
 import humanfriendly
 
 from biomaj_core.utils import Utils
+from biomaj_core.config import BiomajConfig
+
 from biomaj_download.download.interface import DownloadInterface
 
-if sys.version_info[0] < 3:
-    from urllib import urlencode
-else:
-    from urllib.parse import urlencode
-
-try:
-    from io import BytesIO
-except ImportError:
-    from StringIO import StringIO as BytesIO
-
-# We use stat.filemode to convert from mode octal value to string.
-# In python < 3.3, stat.filmode is not defined.
-# This code is copied from the current implementation of stat.filemode.
-if 'filemode' not in stat.__dict__:
-    _filemode_table = (
-        ((stat.S_IFLNK,                "l"),    # noqa: E241
-         (stat.S_IFREG,                "-"),    # noqa: E241
-         (stat.S_IFBLK,                "b"),    # noqa: E241
-         (stat.S_IFDIR,                "d"),    # noqa: E241
-         (stat.S_IFCHR,                "c"),    # noqa: E241
-         (stat.S_IFIFO,                "p")),   # noqa: E241
-        ((stat.S_IRUSR,                "r"),),  # noqa: E241
-        ((stat.S_IWUSR,                "w"),),  # noqa: E241
-        ((stat.S_IXUSR | stat.S_ISUID, "s"),    # noqa: E241
-         (stat.S_ISUID,                "S"),    # noqa: E241
-         (stat.S_IXUSR,                "x")),   # noqa: E241
-        ((stat.S_IRGRP,                "r"),),  # noqa: E241
-        ((stat.S_IWGRP,                "w"),),  # noqa: E241
-        ((stat.S_IXGRP | stat.S_ISGID, "s"),    # noqa: E241
-         (stat.S_ISGID,                "S"),    # noqa: E241
-         (stat.S_IXGRP,                "x")),   # noqa: E241
-        ((stat.S_IROTH,                "r"),),  # noqa: E241
-        ((stat.S_IWOTH,                "w"),),  # noqa: E241
-        ((stat.S_IXOTH | stat.S_ISVTX, "t"),    # noqa: E241
-         (stat.S_ISVTX,                "T"),    # noqa: E241
-         (stat.S_IXOTH,                "x"))    # noqa: E241
-    )
-
-    def _filemode(mode):
-        """Convert a file's mode to a string of the form '-rwxrwxrwx'."""
-        perm = []
-        for table in _filemode_table:
-            for bit, char in table:
-                if mode & bit == bit:
-                    perm.append(char)
-                    break
-            else:
-                perm.append("-")
-        return "".join(perm)
-
-    stat.filemode = _filemode
-
 
 class HTTPParse(object):
 
@@ -113,6 +63,21 @@ class CurlDownload(DownloadInterface):
         ftputil.stat.MSParser(),
     ]
 
+    # Valid values for ftp_method options as string and int
+    VALID_FTP_FILEMETHOD = {
+        "default": pycurl.FTPMETHOD_DEFAULT,
+        "multicwd": pycurl.FTPMETHOD_MULTICWD,
+        "nocwd": pycurl.FTPMETHOD_NOCWD,
+        "singlecwd": pycurl.FTPMETHOD_SINGLECWD,
+    }
+
+    # Valid values for ssh_new_host options as string and int
+    VALID_SSH_NEW_HOST = {
+        "reject": pycurl.KHSTAT_REJECT,
+        "accept": pycurl.KHSTAT_FINE,
+        "add": pycurl.KHSTAT_FINE_ADD_TO_FILE,
+    }
+
     def __init__(self, curl_protocol, host, rootdir, http_parse=None):
         """
         Initialize a CurlDownloader.
@@ -131,6 +96,19 @@ class CurlDownload(DownloadInterface):
         """
         DownloadInterface.__init__(self)
         self.logger.debug('Download')
+
+        # Check for ssh support
+        curl_opts_info = pycurl.version_info()
+        curl_opts = []
+        for opt in curl_opts_info:
+            if isinstance(opt, tuple):
+                for o in opt:
+                    curl_opts.append(o)
+            else:
+                curl_opts.append(opt)
+        if 'sftp' not in curl_opts:
+            CurlDownload.ALL_PROTOCOLS = CurlDownload.FTP_PROTOCOL_FAMILY + CurlDownload.HTTP_PROTOCOL_FAMILY
+            self.logger.warning("sftp not supported by curl: %s" % str(curl_opts_info))
         # Initialize curl_protocol.
         # Note that we don't change that field in set_protocol since this
         # method uses the protocol from the configuration file. It's not clear
@@ -143,15 +121,15 @@ class CurlDownload(DownloadInterface):
         if self.curl_protocol in self.FTP_PROTOCOL_FAMILY:
             self.protocol_family = "ftp"
             self._parse_result = self._ftp_parse_result
-            self.ERRCODE_OK = 226
+            self.ERRCODE_OK = [221, 226]
         elif self.curl_protocol in self.HTTP_PROTOCOL_FAMILY:
             self.protocol_family = "http"
             self._parse_result = self._http_parse_result
-            self.ERRCODE_OK = 200
+            self.ERRCODE_OK = [200]
         elif self.curl_protocol in self.SFTP_PROTOCOL_FAMILY:
             self.protocol_family = "sftp"
             self._parse_result = self._ftp_parse_result
-            self.ERRCODE_OK = 0
+            self.ERRCODE_OK = [0]
         else:  # Should not happen since we check before
             raise ValueError("Unknown protocol")
         self.rootdir = rootdir
@@ -162,7 +140,9 @@ class CurlDownload(DownloadInterface):
         # This object is shared by all operations to use the cache.
         # Before using it, call method:`_basic_curl_configuration`.
         self.crl = pycurl.Curl()
+        #
         # Initialize options
+        #
         # Should we skip SSL verification (cURL -k/--insecure option)
         self.ssl_verifyhost = True
         self.ssl_verifypeer = True
@@ -170,11 +150,34 @@ class CurlDownload(DownloadInterface):
         self.ssl_server_cert = None
         # Keep alive
         self.tcp_keepalive = 0
+        # FTP method (cURL --ftp-method option)
+        self.ftp_method = pycurl.FTPMETHOD_DEFAULT  # Use cURL default
+        # TODO: Don't store default values in BiomajConfig.DEFAULTS for
+        # ssh_hosts_file and ssh_new_hosts
+        # known_hosts file
+        self.ssh_hosts_file = BiomajConfig.DEFAULTS["ssh_hosts_file"]
+        # How to treat unknown host
+        self.ssh_new_host = self.VALID_SSH_NEW_HOST[BiomajConfig.DEFAULTS["ssh_new_host"]]
+        # Allow redirections
+        self.allow_redirections = True
 
-    def _basic_curl_configuration(self):
+    def _accept_new_hosts(self, known_key, found_key, match):
+        # Key found in file: we can accept it
+        # Don't use KHSTAT_FINE_ADD_TO_FILE because the key would be duplicated
+        # See https://github.com/curl/curl/issues/4953.
+        if match == pycurl.KHMATCH_OK:
+            return pycurl.KHSTAT_FINE
+        # Key not found in file: use the ssh_new_host option
+        elif match == pycurl.KHMATCH_MISSING:
+            return self.ssh_new_host
+        # Key missmatch: the best option is to reject it
+        else:
+            return pycurl.KHSTAT_REJECT
+
+    def _network_configuration(self):
         """
         Perform basic configuration (i.e. that doesn't depend on the
-        operation: _download or list). This method shoulmd be called before any
+        operation: _download or list). This method should be called before any
         operation.
         """
         # Reset cURL options before setting them
@@ -188,11 +191,19 @@ class CurlDownload(DownloadInterface):
         if self.credentials is not None:
             self.crl.setopt(pycurl.USERPWD, self.credentials)
 
+        # Hosts file & function to decide for new hosts
+        if self.curl_protocol in self.SFTP_PROTOCOL_FAMILY:
+            self.crl.setopt(pycurl.SSH_KNOWNHOSTS, self.ssh_hosts_file)
+            self.crl.setopt(pycurl.SSH_KEYFUNCTION, self._accept_new_hosts)
+
         # Configure TCP keepalive
         if self.tcp_keepalive:
-            self.crl.setopt(pycurl.TCP_KEEPALIVE, True)
-            self.crl.setopt(pycurl.TCP_KEEPIDLE, self.tcp_keepalive * 2)
-            self.crl.setopt(pycurl.TCP_KEEPINTVL, self.tcp_keepalive)
+            try:
+                self.crl.setopt(pycurl.TCP_KEEPALIVE, True)
+                self.crl.setopt(pycurl.TCP_KEEPIDLE, self.tcp_keepalive * 2)
+                self.crl.setopt(pycurl.TCP_KEEPINTVL, self.tcp_keepalive)
+            except Exception as e:
+                self.logger.exception("TCP keepalive option failed: " + str(e))
 
         # Configure SSL verification (on some platforms, disabling
         # SSL_VERIFYPEER implies disabling SSL_VERIFYHOST so we set
@@ -208,6 +219,12 @@ class CurlDownload(DownloadInterface):
             # CURLOPT_CAPATH is for a directory of certificates.
             self.crl.setopt(pycurl.CAINFO, self.ssl_server_cert)
 
+        # Configure ftp method
+        self.crl.setopt(pycurl.FTP_FILEMETHOD, self.ftp_method)
+
+        # Configure redirections
+        self.crl.setopt(pycurl.FOLLOWLOCATION, self.allow_redirections)
+
         # Configure timeouts
         self.crl.setopt(pycurl.CONNECTTIMEOUT, 300)
         self.crl.setopt(pycurl.TIMEOUT, self.timeout)
@@ -248,16 +265,32 @@ class CurlDownload(DownloadInterface):
         super(CurlDownload, self).set_server(server)
         self.url = self.curl_protocol + '://' + self.server
 
-    def set_options(self, protocol_options):
-        super(CurlDownload, self).set_options(protocol_options)
-        if "ssl_verifyhost" in protocol_options:
-            self.ssl_verifyhost = Utils.to_bool(protocol_options["ssl_verifyhost"])
-        if "ssl_verifypeer" in protocol_options:
-            self.ssl_verifypeer = Utils.to_bool(protocol_options["ssl_verifypeer"])
-        if "ssl_server_cert" in protocol_options:
-            self.ssl_server_cert = protocol_options["ssl_server_cert"]
-        if "tcp_keepalive" in protocol_options:
-            self.tcp_keepalive = Utils.to_int(protocol_options["tcp_keepalive"])
+    def set_options(self, options):
+        super(CurlDownload, self).set_options(options)
+        if "ssl_verifyhost" in options:
+            self.ssl_verifyhost = Utils.to_bool(options["ssl_verifyhost"])
+        if "ssl_verifypeer" in options:
+            self.ssl_verifypeer = Utils.to_bool(options["ssl_verifypeer"])
+        if "ssl_server_cert" in options:
+            self.ssl_server_cert = options["ssl_server_cert"]
+        if "tcp_keepalive" in options:
+            self.tcp_keepalive = Utils.to_int(options["tcp_keepalive"])
+        if "ftp_method" in options:
+            # raw_val is a string which contains the name of the option as in the CLI.
+            # We always convert raw_val to a valid integer
+            raw_val = options["ftp_method"].lower()
+            if raw_val not in self.VALID_FTP_FILEMETHOD:
+                raise ValueError("Invalid value for ftp_method")
+            self.ftp_method = self.VALID_FTP_FILEMETHOD[raw_val]
+        if "ssh_hosts_file" in options:
+            self.ssh_hosts_file = options["ssh_hosts_file"]
+        if "ssh_new_host" in options:
+            raw_val = options["ssh_new_host"].lower()
+            if raw_val not in self.VALID_SSH_NEW_HOST:
+                raise ValueError("Invalid value for ssh_new_host")
+            self.ssh_new_host = self.VALID_SSH_NEW_HOST[raw_val]
+        if "allow_redirections" in options:
+            self.allow_redirections = Utils.to_bool(options["allow_redirections"])
 
     def _append_file_to_download(self, rfile):
         # Add url and root to the file if needed (for safety)
@@ -270,66 +303,74 @@ class CurlDownload(DownloadInterface):
     def _file_url(self, rfile):
         # rfile['root'] is set to self.rootdir if needed but may be different.
         # We don't use os.path.join because rfile['name'] may starts with /
-        return self.url + '/' + rfile['root'] + rfile['name']
+        url = self.url + '/' + rfile['root'] + rfile['name']
+        url_elts = url.split('://')
+        if len(url_elts) == 2:
+            url_elts[1] = re.sub("/{2,}", "/", url_elts[1])
+            return '://'.join(url_elts)
+        return re.sub("/{2,}", "/", url)
 
     def _download(self, file_path, rfile):
         """
+        Download one file and return False in case of success and True
+        otherwise.
+
         This method is designed to work for FTP(S), HTTP(S) and SFTP.
         """
         error = True
-        nbtry = 1
         # Forge URL of remote file
         file_url = self._file_url(rfile)
-        while(error is True and nbtry < 3):
 
-            self._basic_curl_configuration()
+        try:
+            self.crl.setopt(pycurl.URL, file_url)
+        except Exception:
+            self.crl.setopt(pycurl.URL, file_url.encode('ascii', 'ignore'))
 
-            try:
-                self.crl.setopt(pycurl.URL, file_url)
-            except Exception:
-                self.crl.setopt(pycurl.URL, file_url.encode('ascii', 'ignore'))
+        # Create file and assign it to the pycurl object
+        fp = open(file_path, "wb")
+        self.crl.setopt(pycurl.WRITEFUNCTION, fp.write)
 
-            # Create file and assign it to the pycurl object
-            fp = open(file_path, "wb")
-            self.crl.setopt(pycurl.WRITEFUNCTION, fp.write)
+        # This is specific to HTTP
+        if self.method == 'POST':
+            # Form data must be provided already urlencoded.
+            postfields = urlencode(self.param)
+            # Sets request method to POST,
+            # Content-Type header to application/x-www-form-urlencoded
+            # and data to send in request body.
+            self.crl.setopt(pycurl.POSTFIELDS, postfields)
 
-            # This is specific to HTTP
-            if self.method == 'POST':
-                # Form data must be provided already urlencoded.
-                postfields = urlencode(self.param)
-                # Sets request method to POST,
-                # Content-Type header to application/x-www-form-urlencoded
-                # and data to send in request body.
-                self.crl.setopt(pycurl.POSTFIELDS, postfields)
+        # Try download (we don't raise errors here since its the return value
+        # ('error') that matters for the calling method; this is set to True
+        # only in case of success).
+        try:
+            self.crl.perform()
+            errcode = self.crl.getinfo(pycurl.RESPONSE_CODE)
+            if int(errcode) not in self.ERRCODE_OK:
+                error = True
+                self.logger.error('Error while downloading ' + file_url + ' - ' + str(errcode))
+            else:
+                error = False
+        except Exception as e:
+            self.logger.error('Error while downloading ' + file_url + ' - ' + str(e))
 
-            # Try download
-            try:
-                self.crl.perform()
-                errcode = self.crl.getinfo(pycurl.RESPONSE_CODE)
-                if int(errcode) != self.ERRCODE_OK:
-                    error = True
-                    self.logger.error('Error while downloading ' + file_url + ' - ' + str(errcode))
-                else:
-                    error = False
-            except Exception as e:
-                self.logger.error('Could not get errcode:' + str(e))
+        # Check if we were redirected
+        if self.curl_protocol in self.HTTP_PROTOCOL_FAMILY:
+            n_redirect = self.crl.getinfo(pycurl.REDIRECT_COUNT)
+            if n_redirect:
+                real_url = self.crl.getinfo(pycurl.EFFECTIVE_URL)
+                redirect_time = self.crl.getinfo(pycurl.REDIRECT_TIME)
+                msg_fmt = 'Download was redirected to %s (%i redirection(s), took %ss)'
+                msg = msg_fmt % (real_url, n_redirect, redirect_time)
+                self.logger.info(msg)
 
-            # Close file
-            fp.close()
+        # Close file
+        fp.close()
 
-            # Check that the archive is correct
-            if not error and not self.skip_check_uncompress:
-                archive_status = Utils.archive_check(file_path)
-                if not archive_status:
-                    self.logger.error('Archive is invalid or corrupted, deleting file and retrying download')
-                    error = True
-                    if os.path.exists(file_path):
-                        os.remove(file_path)
+        if error:
+            return error
 
-            # Increment retry counter
-            nbtry += 1
-
-        return error
+        # Our part is done so call parent _download
+        return super(CurlDownload, self)._download(file_path, rfile)
 
     def list(self, directory=''):
         '''
@@ -340,10 +381,11 @@ class CurlDownload(DownloadInterface):
         This is a generic method for HTTP and FTP. The protocol-specific parts
         are done in _<protocol>_parse_result.
         '''
-        dir_url = self.url + self.rootdir + directory
+        dirbase = re.sub('//+', '/', self.rootdir + directory)
+        dir_url = self.url + dirbase
         self.logger.debug('Download:List:' + dir_url)
 
-        self._basic_curl_configuration()
+        self._network_configuration()
 
         try:
             self.crl.setopt(pycurl.URL, dir_url)
@@ -357,8 +399,25 @@ class CurlDownload(DownloadInterface):
         # Try to list
         try:
             self.crl.perform()
+            errcode = self.crl.getinfo(pycurl.RESPONSE_CODE)
+            if int(errcode) not in self.ERRCODE_OK:
+                msg = 'Error while listing ' + dir_url + ' - ' + str(errcode)
+                self.logger.error(msg)
+                raise Exception(msg)
         except Exception as e:
-            self.logger.error('Could not get errcode:' + str(e))
+            msg = 'Error while listing ' + dir_url + ' - ' + str(e)
+            self.logger.error(msg)
+            raise e
+
+        # Check if we were redirected
+        if self.curl_protocol in self.HTTP_PROTOCOL_FAMILY:
+            n_redirect = self.crl.getinfo(pycurl.REDIRECT_COUNT)
+            if n_redirect:
+                real_url = self.crl.getinfo(pycurl.EFFECTIVE_URL)
+                redirect_time = self.crl.getinfo(pycurl.REDIRECT_TIME)
+                msg_fmt = 'Download was redirected to %s (%i redirection(s), took %ss)'
+                msg = msg_fmt % (real_url, n_redirect, redirect_time)
+                self.logger.info(msg)
 
         # Figure out what encoding was sent with the response, if any.
         # Check against lowercased header name.
