@@ -1,5 +1,9 @@
 """
 Note that attributes 'network' and 'local_irods' are ignored for CI.
+
+To run 'local_irods' tests, you need an iRODS server on localhost (default port,
+user 'rods', password 'rods') and a zone /tempZone/home/rods. See
+UtilsForLocalIRODSTest.
 """
 from nose.plugins.attrib import attr
 
@@ -12,8 +16,11 @@ import stat
 
 from mock import patch
 
+from irods.session import iRODSSession
+
 from biomaj_core.config import BiomajConfig
 from biomaj_core.utils import Utils
+from biomaj_download.download.interface import DownloadInterface
 from biomaj_download.download.curl import CurlDownload, HTTPParse
 from biomaj_download.download.direct import DirectFTPDownload, DirectHTTPDownload
 from biomaj_download.download.localcopy  import LocalDownload
@@ -21,6 +28,8 @@ from biomaj_download.download.rsync import RSYNCDownload
 from biomaj_download.download.protocolirods import IRODSDownload
 
 import unittest
+import tenacity
+
 
 class UtilsForTest():
   """
@@ -63,6 +72,11 @@ class UtilsForTest():
     if self.bank_properties is None:
       self.__copy_test_bank_properties()
 
+    # Create an invalid archive file (empty file). This is deleted by clean().
+    # See TestBiomajRSYNCDownload.test_rsync_download_skip_check_uncompress.
+    self.invalid_archive = os.path.join(self.test_dir, 'invalid.gz')
+    open(self.invalid_archive, 'w').close()
+
   def clean(self):
     """
     Deletes temp directory
@@ -72,13 +86,15 @@ class UtilsForTest():
   def __copy_test_bank_properties(self):
     if self.bank_properties is not None:
       return
-    self.bank_properties = ['alu', 'local', 'testhttp','directhttp']
+    # Copy bank configuration (those bank use external resources so there is no tuning to do)
+    self.bank_properties = ['alu', 'testhttp', 'directhttp', 'multi']
     curdir = os.path.dirname(os.path.realpath(__file__))
     for b in self.bank_properties:
         from_file = os.path.join(curdir, b+'.properties')
         to_file = os.path.join(self.conf_dir, b+'.properties')
         shutil.copyfile(from_file, to_file)
 
+    # Copy bank process
     self.bank_process = ['test.sh']
     curdir = os.path.dirname(os.path.realpath(__file__))
     procdir = os.path.join(curdir, 'bank/process')
@@ -88,11 +104,11 @@ class UtilsForTest():
       shutil.copyfile(from_file, to_file)
       os.chmod(to_file, stat.S_IRWXU)
 
-    # Manage local bank test, use bank test subdir as remote
-    properties = ['multi.properties', 'computederror.properties', 'error.properties', 'local.properties', 'localprocess.properties', 'testhttp.properties', 'computed.properties', 'computed2.properties', 'sub1.properties', 'sub2.properties']
+    # Copy and adapt bank configuration that use local resources: we use the "bank" dir in current test directory as remote
+    properties = ['local', 'localprocess', 'computed', 'computed2', 'sub1', 'sub2', 'computederror', 'error']
     for prop in properties:
-      from_file = os.path.join(curdir, prop)
-      to_file = os.path.join(self.conf_dir, prop)
+      from_file = os.path.join(curdir, prop+'.properties')
+      to_file = os.path.join(self.conf_dir, prop+'.properties')
       fout = open(to_file,'w')
       with open(from_file,'r') as fin:
         for line in fin:
@@ -111,6 +127,7 @@ class UtilsForTest():
     curdir = os.path.dirname(os.path.realpath(__file__))
     global_template = os.path.join(curdir,'global.properties')
     fout = open(self.global_properties,'w')
+    # Adapt directories in global configuration to the current test directory
     with open(global_template,'r') as fin:
         for line in fin:
           if line.startswith('conf.dir'):
@@ -128,44 +145,79 @@ class UtilsForTest():
     fout.close()
 
 
-class TestBiomajUtils(unittest.TestCase):
+class UtilsForLocalIRODSTest(UtilsForTest):
+    """
+    This class is used to prepare 'local_irods' tests.
+    """
+    SERVER = "localhost"
+    PORT = 1247
+    ZONE = "tempZone"
+    USER = "rods"
+    PASSWORD = "rods"
+    COLLECTION = os.path.join("/" + ZONE, "home/rods/")  # Don't remove or add /
 
-  def setUp(self):
-    self.utils = UtilsForTest()
+    def __init__(self):
+        super(UtilsForLocalIRODSTest, self).__init__()
+        self._session = iRODSSession(host=self.SERVER, port=self.PORT,
+                                     user=self.USER, password=self.PASSWORD,
+                                     zone=self.ZONE)
+        self.curdir = os.path.dirname(os.path.realpath(__file__))
+        # Copy some valid archives (bank/test.fasta.gz)
+        file_ = os.path.join(self.curdir, "bank/test.fasta.gz")
+        self._session.data_objects.put(file_, self.COLLECTION)
+        # Copy invalid.gz
+        self._session.data_objects.put(self.invalid_archive, self.COLLECTION)
 
-  def tearDown(self):
-    self.utils.clean()
+    def clean(self):
+        super(UtilsForLocalIRODSTest, self).clean()
+        # Remove files on iRODS (use force otherwise the files are put in trash)
+        # Remove test.fasta.gz
+        self._session.data_objects.unlink(os.path.join(self.COLLECTION, "test.fasta.gz"), force=True)
+        # Remove invalid.gz
+        self._session.data_objects.unlink(os.path.join(self.COLLECTION, "invalid.gz"), force=True)
 
 
-  def test_mimes(self):
-    fasta_file = os.path.join(os.path.dirname(os.path.realpath(__file__)),'bank/test2.fasta')
-    (mime, encoding) = Utils.detect_format(fasta_file)
-    self.assertTrue('application/fasta' == mime)
+class TestDownloadInterface(unittest.TestCase):
+  """
+  Test of the interface.
+  """
 
-  @attr('compress')
-  def test_uncompress(self):
-    from_file = { 'root': os.path.dirname(os.path.realpath(__file__)),
-                  'name': 'bank/test.fasta.gz'
-                  }
+  def test_retry_parsing(self):
+    """
+    Test parsing of stop and wait conditions.
+    """
+    downloader = DownloadInterface()
+    # Test some garbage
+    d = dict(stop_condition="stop_after_attempts")  # no param
+    self.assertRaises(ValueError, downloader.set_options, d)
+    d = dict(stop_condition="1 & 1")  # not a stop_condition
+    self.assertRaises(ValueError, downloader.set_options, d)
+    d = dict(stop_condition="stop_after_attempts(5) & 1")  # not a stop_condition
+    self.assertRaises(ValueError, downloader.set_options, d)
+    # Test some garbage
+    d = dict(wait_policy="wait_random")  # no param
+    self.assertRaises(ValueError, downloader.set_options, d)
+    d = dict(wait_policy="I love python")  # not a wait_condition
+    self.assertRaises(ValueError, downloader.set_options, d)
+    d = dict(wait_policy="wait_random(5) + 3")  # not a wait_condition
+    self.assertRaises(ValueError, downloader.set_options, d)
+    # Test operators
+    d = dict(stop_condition="stop_never | stop_after_attempt(5)",
+             wait_policy="wait_none + wait_random(1, 2)")
+    downloader.set_options(d)
+    # Test wait_combine, wait_chain
+    d = dict(wait_policy="wait_combine(wait_fixed(3), wait_random(1, 2))")
+    downloader.set_options(d)
+    d = dict(wait_policy="wait_chain(wait_fixed(3), wait_random(1, 2))")
+    downloader.set_options(d)
+    # Test stop_any and stop_all
+    stop_condition = "stop_any(stop_after_attempt(5), stop_after_delay(10))"
+    d = dict(stop_condition=stop_condition)
+    downloader.set_options(d)
+    stop_condition = "stop_all(stop_after_attempt(5), stop_after_delay(10))"
+    d = dict(stop_condition=stop_condition)
+    downloader.set_options(d)
 
-    to_dir = self.utils.data_dir
-    Utils.copy_files([from_file], to_dir)
-    Utils.uncompress(os.path.join(to_dir, from_file['name']))
-    self.assertTrue(os.path.exists(to_dir+'/bank/test.fasta'))
-
-  def test_copy_with_regexp(self):
-    from_dir = os.path.dirname(os.path.realpath(__file__))
-    to_dir = self.utils.data_dir
-    Utils.copy_files_with_regexp(from_dir, to_dir, ['.*\.py'])
-    self.assertTrue(os.path.exists(to_dir+'/biomaj_tests.py'))
-
-  def test_copy(self):
-    from_dir = os.path.dirname(os.path.realpath(__file__))
-    local_file = 'biomaj_tests.py'
-    files_to_copy = [ {'root': from_dir, 'name': local_file}]
-    to_dir = self.utils.data_dir
-    Utils.copy_files(files_to_copy, to_dir)
-    self.assertTrue(os.path.exists(to_dir+'/biomaj_tests.py'))
 
 class TestBiomajLocalDownload(unittest.TestCase):
   """
@@ -190,6 +242,16 @@ class TestBiomajLocalDownload(unittest.TestCase):
     locald.close()
     self.assertTrue(len(file_list) > 1)
 
+  def test_local_list_error(self):
+    locald = LocalDownload("/tmp/foo/")
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = locald.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+    locald.close()
+
   def test_local_download(self):
     locald = LocalDownload(self.examples)
     (file_list, dir_list) = locald.list()
@@ -230,7 +292,8 @@ class TestBiomajLocalDownload(unittest.TestCase):
     except Exception:
       msg = "In %s: copy worked but hardlinks were not used." % self.id()
       logging.info(msg)
-      
+
+
 @attr('network')
 @attr('http')
 class TestBiomajHTTPDownload(unittest.TestCase):
@@ -240,14 +303,16 @@ class TestBiomajHTTPDownload(unittest.TestCase):
   def setUp(self):
     self.utils = UtilsForTest()
     BiomajConfig.load_config(self.utils.global_properties, allow_user_config=False)
+    # Create an HTTPParse object used for most tests from the config file testhttp
     self.config = BiomajConfig('testhttp')
-    self.http_parse = HTTPParse(self.config.get('http.parse.dir.line'),
+    self.http_parse = HTTPParse(
+        self.config.get('http.parse.dir.line'),
         self.config.get('http.parse.file.line'),
         int(self.config.get('http.group.dir.name')),
         int(self.config.get('http.group.dir.date')),
         int(self.config.get('http.group.file.name')),
         int(self.config.get('http.group.file.date')),
-        self.config.get('http.group.file.date_format', None),
+        self.config.get('http.group.file.date_format'),
         int(self.config.get('http.group.file.size'))
     )
 
@@ -260,26 +325,37 @@ class TestBiomajHTTPDownload(unittest.TestCase):
     httpd.close()
     self.assertTrue(len(file_list) == 1)
 
+  def test_http_list_error(self):
+    """
+    Test that errors in list are correctly caught.
+    """
+    # Test access to non-existent directory
+    httpd = CurlDownload('http', 'ftp2.fr.debian.org', '/debian/dists/foo/', self.http_parse)
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = httpd.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+
   def test_http_list_dateregexp(self):
-    #self.http_parse.file_date_format = "%%d-%%b-%%Y %%H:%%M"
-    self.http_parse.file_date_format = "%%Y-%%m-%%d %%H:%%M"
     httpd = CurlDownload('http', 'ftp2.fr.debian.org', '/debian/dists/', self.http_parse)
     (file_list, dir_list) = httpd.list()
     httpd.close()
     self.assertTrue(len(file_list) == 1)
 
   def test_http_download_no_size(self):
-    self.http_parse = HTTPParse(self.config.get('http.parse.dir.line'),
+    # Create a custom http_parse without size
+    http_parse = HTTPParse(self.config.get('http.parse.dir.line'),
         self.config.get('http.parse.file.line'),
         int(self.config.get('http.group.dir.name')),
         int(self.config.get('http.group.dir.date')),
         int(self.config.get('http.group.file.name')),
         int(self.config.get('http.group.file.date')),
-        self.config.get('http.group.file.date_format', None),
+        self.config.get('http.group.file.date_format'),
         -1
     )
-    self.http_parse.file_date_format = "%%Y-%%m-%%d %%H:%%M"
-    httpd = CurlDownload('http', 'ftp2.fr.debian.org', '/debian/dists/', self.http_parse)
+    httpd = CurlDownload('http', 'ftp2.fr.debian.org', '/debian/dists/', http_parse)
     (file_list, dir_list) = httpd.list()
     httpd.match([r'^README$'], file_list, dir_list)
     httpd.download(self.utils.data_dir)
@@ -287,16 +363,17 @@ class TestBiomajHTTPDownload(unittest.TestCase):
     self.assertTrue(len(httpd.files_to_download) == 1)
 
   def test_http_download_no_date(self):
-    self.http_parse = HTTPParse(self.config.get('http.parse.dir.line'),
+    # Create a custom http_parse without date
+    http_parse = HTTPParse(self.config.get('http.parse.dir.line'),
         self.config.get('http.parse.file.line'),
         int(self.config.get('http.group.dir.name')),
         int(self.config.get('http.group.dir.date')),
         int(self.config.get('http.group.file.name')),
         -1,
-        self.config.get('http.group.file.date_format', None),
+        None,
         int(self.config.get('http.group.file.size'))
     )
-    httpd = CurlDownload('http', 'ftp2.fr.debian.org', '/debian/dists/', self.http_parse)
+    httpd = CurlDownload('http', 'ftp2.fr.debian.org', '/debian/dists/', http_parse)
     (file_list, dir_list) = httpd.list()
     httpd.match([r'^README$'], file_list, dir_list)
     httpd.download(self.utils.data_dir)
@@ -304,7 +381,6 @@ class TestBiomajHTTPDownload(unittest.TestCase):
     self.assertTrue(len(httpd.files_to_download) == 1)
 
   def test_http_download(self):
-    self.http_parse.file_date_format = "%%Y-%%m-%%d %%H:%%M"
     httpd = CurlDownload('http', 'ftp2.fr.debian.org', '/debian/dists/', self.http_parse)
     (file_list, dir_list) = httpd.list()
     print(str(file_list))
@@ -314,7 +390,6 @@ class TestBiomajHTTPDownload(unittest.TestCase):
     self.assertTrue(len(httpd.files_to_download) == 1)
 
   def test_http_download_in_subdir(self):
-    self.http_parse.file_date_format = "%%Y-%%m-%%d %%H:%%M"
     httpd = CurlDownload('http', 'ftp2.fr.debian.org', '/debian/', self.http_parse)
     (file_list, dir_list) = httpd.list()
     httpd.match([r'^dists/README$'], file_list, dir_list)
@@ -322,6 +397,41 @@ class TestBiomajHTTPDownload(unittest.TestCase):
     httpd.close()
     self.assertTrue(len(httpd.files_to_download) == 1)
 
+  def test_redirection(self):
+    """
+    Test HTTP redirections
+    """
+    # The site used in this test redirects to https (see #33).
+    http_parse = HTTPParse(
+        r'<img[\s]+src="[\S]+"[\s]+alt="\[[\s]+\]"[\s]*/?>[\s]<a[\s]+href="([\S]+)".*([\d]{4}-[\d]{2}-[\d]{2}\s[\d]{2}:[\d]{2})[\s]+-',
+        r'<img[\s]+src="[\S]+"[\s]+alt="\[[\s]+\]"[\s]*/?>[\s]<a[\s]+href="([\S]+)".*([\d]{4}-[\d]{2}-[\d]{2}\s[\d]{2}:[\d]{2})[\s]+([\d]+)',
+        1,
+        2,
+        1,
+        2,
+        "%%Y-%%m-%%d %%H:%%M",
+        3
+    )
+    # First test: allow redirections
+    httpd = CurlDownload('http', 'plasmodb.org', '/common/downloads/Current_Release/', http_parse)
+    (file_list, dir_list) = httpd.list()
+    httpd.match([r'^Build_number$'], file_list, dir_list)
+    # Check that we have been redirected to HTTPS by inspecting logs
+    with self.assertLogs(logger="biomaj", level="INFO") as cm:
+      httpd.download(self.utils.data_dir)
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Download was redirected to https://")
+    httpd.close()
+    self.assertTrue(len(httpd.files_to_download) == 1)
+    # Second test: disable redirections (hence listing fails)
+    httpd = CurlDownload('http', 'plasmodb.org', '/common/downloads/Current_Release/', http_parse)
+    httpd.set_options({
+      "allow_redirections": False
+    })
+    with self.assertRaises(Exception):
+      (file_list, dir_list) = httpd.list()
+    httpd.close()
+
 
 @attr('network')
 @attr('https')
@@ -337,19 +447,17 @@ class TestBiomajHTTPSDownload(unittest.TestCase):
     self.utils.clean()
 
   def test_download(self):
-    self.utils = UtilsForTest()
-    self.http_parse = HTTPParse(
+    http_parse = HTTPParse(
         "<a[\s]+href=\"([\w\-\.]+\">[\w\-\.]+.tar.gz)<\/a>[\s]+([0-9]{2}-[A-Za-z]{3}-[0-9]{4}[\s][0-9]{2}:[0-9]{2})[\s]+([0-9]+[A-Za-z])",
         "<a[\s]+href=\"[\w\-\.]+\">([\w\-\.]+.tar.gz)<\/a>[\s]+([0-9]{2}-[A-Za-z]{3}-[0-9]{4}[\s][0-9]{2}:[0-9]{2})[\s]+([0-9]+[A-Za-z])",
         1,
         2,
         1,
         2,
-        None,
+        "%%d-%%b-%%Y %%H:%%M",
         3
     )
-    self.http_parse.file_date_format = "%%d-%%b-%%Y %%H:%%M"
-    httpd = CurlDownload('https', 'mirrors.edge.kernel.org', '/pub/software/scm/git/debian/', self.http_parse)
+    httpd = CurlDownload('https', 'mirrors.edge.kernel.org', '/pub/software/scm/git/debian/', http_parse)
     (file_list, dir_list) = httpd.list()
     httpd.match([r'^git-core-0.99.6.tar.gz$'], file_list, dir_list)
     httpd.download(self.utils.data_dir)
@@ -364,17 +472,48 @@ class TestBiomajSFTPDownload(unittest.TestCase):
   Test SFTP downloader
   """
 
-  PROTOCOL = "ftps"
-  
+  PROTOCOL = "sftp"
+
   def setUp(self):
     self.utils = UtilsForTest()
+    # Temporary host key file in test dir (so this is cleaned)
+    (_, self.khfile) = tempfile.mkstemp(dir=self.utils.test_dir)
 
   def tearDown(self):
     self.utils.clean()
 
+  def test_list_error(self):
+    """
+    Test that errors in list are correctly caught.
+    """
+    # Test access to non-existent directory
+    sftpd = CurlDownload(self.PROTOCOL, "test.rebex.net", "/toto")
+    sftpd.set_credentials("demo:password")
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = sftpd.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+    sftpd.close()
+    # Test with wrong password
+    sftpd = CurlDownload(self.PROTOCOL, "test.rebex.net", "/")
+    sftpd.set_credentials("demo:badpassword")
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = sftpd.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+    sftpd.close()
+
   def test_download(self):
     sftpd = CurlDownload(self.PROTOCOL, "test.rebex.net", "/")
     sftpd.set_credentials("demo:password")
+    sftpd.set_options({
+        "ssh_hosts_file": self.khfile,
+        "ssh_new_host": "add"
+    })
     (file_list, dir_list) = sftpd.list()
     sftpd.match([r'^readme.txt$'], file_list, dir_list)
     sftpd.download(self.utils.data_dir)
@@ -403,6 +542,22 @@ class TestBiomajDirectFTPDownload(unittest.TestCase):
     ftpd.close()
     self.assertTrue(len(file_list) == 1)
 
+  def test_ftp_list_error(self):
+    """
+    Test that errors in list are correctly caught.
+    """
+    # Test access to non-existent directory
+    file_list = ['/toto/debian/doc/mailing-lists.txt']
+    ftpd = DirectFTPDownload('ftp', 'ftp.fr.debian.org', '')
+    ftpd.set_files_to_download(file_list)
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = ftpd.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+    ftpd.close()
+
   def test_download(self):
     file_list = ['/debian/doc/mailing-lists.txt']
     ftpd = DirectFTPDownload('ftp', 'ftp.fr.debian.org', '')
@@ -413,14 +568,13 @@ class TestBiomajDirectFTPDownload(unittest.TestCase):
     self.assertTrue(os.path.exists(os.path.join(self.utils.data_dir,'mailing-lists.txt')))
 
 
-
 @attr('directftps')
 @attr('network')
 class TestBiomajDirectFTPSDownload(unittest.TestCase):
   """
   Test DirectFTP downloader with FTPS.
   """
-  
+
   def setUp(self):
     self.utils = UtilsForTest()
 
@@ -473,6 +627,22 @@ class TestBiomajDirectHTTPDownload(unittest.TestCase):
     self.assertTrue(file_list[0]['size']!=0)
     self.assertFalse(fyear == ftpd.files_to_download[0]['year'] and fmonth == ftpd.files_to_download[0]['month'] and fday == ftpd.files_to_download[0]['day'])
 
+  def test_http_list_error(self):
+    """
+    Test that errors in list are correctly caught.
+    """
+    # Test access to non-existent directory
+    file_list = ['/toto/debian/README.html']
+    ftpd = DirectHTTPDownload('http', 'ftp2.fr.debian.org', '')
+    ftpd.set_files_to_download(file_list)
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = ftpd.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+    ftpd.close()
+
   def test_download(self):
     file_list = ['/debian/README.html']
     ftpd = DirectHTTPDownload('http', 'ftp2.fr.debian.org', '')
@@ -497,7 +667,6 @@ class TestBiomajDirectHTTPDownload(unittest.TestCase):
       my_json = json.loads(content)
       self.assertTrue(my_json['args']['key1'] == 'value1')
 
-  @attr('test')
   def test_download_save_as(self):
     file_list = ['/debian/README.html']
     ftpd = DirectHTTPDownload('http', 'ftp2.fr.debian.org', '')
@@ -524,6 +693,32 @@ class TestBiomajDirectHTTPDownload(unittest.TestCase):
       content = content_file.read()
       my_json = json.loads(content)
       self.assertTrue(my_json['form']['key1'] == 'value1')
+      
+  def test_redirection(self):
+    """
+    Test HTTP redirections
+    """
+    # The site used in this test redirects to https (see #33).
+    # First test: allow redirections
+    httpd = DirectHTTPDownload('http', 'plasmodb.org', '/common/downloads/Current_Release/')
+    httpd.set_files_to_download(['Build_number'])
+    httpd.download(self.utils.data_dir)
+    # Check that we have been redirected to HTTPS by inspecting logs
+    with self.assertLogs(logger="biomaj", level="INFO") as cm:
+      httpd.download(self.utils.data_dir)
+       # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Download was redirected to https://")
+    httpd.close()
+    self.assertTrue(len(httpd.files_to_download) == 1)
+    # Second test: block redirections
+    httpd = DirectHTTPDownload('http', 'plasmodb.org', '/common/downloads/Current_Release/')
+    httpd.set_files_to_download(['Build_number'])
+    httpd.set_options({
+      "allow_redirections": False
+    })
+    with self.assertRaises(Exception):
+      httpd.download(self.utils.data_dir)
+    httpd.close()
 
 
 @attr('ftp')
@@ -545,7 +740,31 @@ class TestBiomajFTPDownload(unittest.TestCase):
     ftpd.close()
     self.assertTrue(len(file_list) > 1)
 
-  @attr('test')
+  def test_ftp_list_error(self):
+    """
+    Test that errors in list are correctly caught.
+    """
+    # Test access to non-existent directory
+    ftpd = CurlDownload("ftp", "test.rebex.net", "/toto")
+    ftpd.set_credentials("demo:password")
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = ftpd.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+    ftpd.close()
+    # Test with wrong password
+    ftpd = CurlDownload("ftp", "test.rebex.net", "/")
+    ftpd.set_credentials("demo:badpassword")
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = ftpd.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+    ftpd.close()
+
   def test_download(self):
     ftpd = CurlDownload('ftp', 'speedtest.tele2.net', '/')
     (file_list, dir_list) = ftpd.list()
@@ -604,6 +823,21 @@ class TestBiomajFTPDownload(unittest.TestCase):
     self.assertTrue(len(ftpd.files_to_download)==2)
     self.assertTrue(len(ftpd.files_to_copy)==2)
 
+  @attr('test')
+  def test_download_or_copy_directhttp(self):
+    ftpd = DirectHTTPDownload('https', 'ftp.fr.debian.org', '/debian/')
+    ftpd.files_to_download = [
+          {'name':'/test1', 'year': '2013', 'month': '11', 'day': '10', 'size': 10},
+    ]
+    available_files = [
+        {'name':'/test1', 'year': '2020', 'month': '11', 'day': '10', 'size': 10},
+      # {"root": "/", "permissions": "", "group": "", "user": "", "size": 23723408, "month": 6, "day": 19, "year": 2018, "name": "/common/downloads/release-38/Pfalciparum3D7/fasta/data/PlasmoDB-38_Pfalciparum3D7_Genome.fasta", "hash": "e58669a71eacff7a9dcceed04a8ecdd1", "save_as": "PlasmoDB-38_Pfalciparum3D7_Genome.fasta", "url": "https://plasmodb.org"}
+    ]
+    ftpd.download_or_copy(available_files, '/biomaj', False)
+    ftpd.close()
+    self.assertTrue(len(ftpd.files_to_download)==1)
+    self.assertTrue(len(ftpd.files_to_copy)==0)
+
   def test_get_more_recent_file(self):
     files = [
           {'name':'/test1', 'year': '2013', 'month': '11', 'day': '10', 'size': 10},
@@ -616,14 +850,47 @@ class TestBiomajFTPDownload(unittest.TestCase):
     self.assertTrue(release['month']=='11')
     self.assertTrue(release['day']=='12')
 
+  def test_download_retry(self):
+    """
+    Try to download fake files to test retry.
+    """
+    n_attempts = 5
+    ftpd = CurlDownload("ftp", "speedtest.tele2.net", "/")
+    # Download a fake file
+    ftpd.set_files_to_download([
+          {'name': 'TOTO.zip', 'year': '2016', 'month': '02', 'day': '19',
+           'size': 1, 'save_as': 'TOTO1KB'}
+    ])
+    ftpd.set_options(dict(stop_condition=tenacity.stop.stop_after_attempt(n_attempts),
+                          wait_condition=tenacity.wait.wait_none()))
+    self.assertRaisesRegex(
+        Exception, "^CurlDownload:Download:Error:",
+        ftpd.download, self.utils.data_dir,
+    )
+    logging.debug(ftpd.retryer.statistics)
+    self.assertTrue(len(ftpd.files_to_download) == 1)
+    self.assertTrue(ftpd.retryer.statistics["attempt_number"] == n_attempts)
+    # Try to download another file to ensure that it retryies
+    ftpd.set_files_to_download([
+          {'name': 'TITI.zip', 'year': '2016', 'month': '02', 'day': '19',
+           'size': 1, 'save_as': 'TOTO1KB'}
+    ])
+    self.assertRaisesRegex(
+        Exception, "^CurlDownload:Download:Error:",
+        ftpd.download, self.utils.data_dir,
+    )
+    self.assertTrue(len(ftpd.files_to_download) == 1)
+    self.assertTrue(ftpd.retryer.statistics["attempt_number"] == n_attempts)
+    ftpd.close()
+
   def test_ms_server(self):
-      ftpd = CurlDownload("ftp", "test.rebex.net", "/")
-      ftpd.set_credentials("demo:password")
-      (file_list, dir_list) = ftpd.list()
-      ftpd.match(["^readme.txt$"], file_list, dir_list)
-      ftpd.download(self.utils.data_dir)
-      ftpd.close()
-      self.assertTrue(len(ftpd.files_to_download) == 1)
+    ftpd = CurlDownload("ftp", "test.rebex.net", "/")
+    ftpd.set_credentials("demo:password")
+    (file_list, dir_list) = ftpd.list()
+    ftpd.match(["^readme.txt$"], file_list, dir_list)
+    ftpd.download(self.utils.data_dir)
+    ftpd.close()
+    self.assertTrue(len(ftpd.files_to_download) == 1)
 
   def test_download_tcp_keepalive(self):
       """
@@ -639,6 +906,20 @@ class TestBiomajFTPDownload(unittest.TestCase):
       ftpd.close()
       self.assertTrue(len(ftpd.files_to_download) == 1)
 
+  def test_download_ftp_method(self):
+      """
+      Test setting ftp_method (it probably doesn't change anything here but we
+      test that there is no obvious mistake in the code).
+      """
+      ftpd = CurlDownload("ftp", "test.rebex.net", "/")
+      ftpd.set_options(dict(ftp_method="nocwd"))
+      ftpd.set_credentials("demo:password")
+      (file_list, dir_list) = ftpd.list()
+      ftpd.match(["^readme.txt$"], file_list, dir_list)
+      ftpd.download(self.utils.data_dir)
+      ftpd.close()
+      self.assertTrue(len(ftpd.files_to_download) == 1)
+
 
 @attr('ftps')
 @attr('network')
@@ -661,6 +942,31 @@ class TestBiomajFTPSDownload(unittest.TestCase):
     ftpd.close()
     self.assertTrue(len(file_list) == 1)
 
+  def test_ftps_list_error(self):
+    """
+    Test that errors in list are correctly caught.
+    """
+    # Test access to non-existent directory
+    ftpd = CurlDownload("ftps", "test.rebex.net", "/toto")
+    ftpd.set_credentials("demo:password")
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = ftpd.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+    ftpd.close()
+    # Test with wrong password
+    ftpd = CurlDownload("ftps", "test.rebex.net", "/")
+    ftpd.set_credentials("demo:badpassword")
+    # Check that we raise an exception and log a message
+    with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+      with self.assertRaises(Exception):
+        (file_list, dir_list) = ftpd.list()
+      # Test log message format (we assume that there is only 1 message)
+      self.assertRegex(cm.output[0], "Error while listing")
+    ftpd.close()
+
   def test_download(self):
     ftpd = CurlDownload(self.PROTOCOL, "test.rebex.net", "/")
     ftpd.set_credentials("demo:password")
@@ -674,7 +980,7 @@ class TestBiomajFTPSDownload(unittest.TestCase):
     # This server is misconfigured hence we disable all SSL verification
     SERVER = "demo.wftpserver.com"
     DIRECTORY = "/download/"
-    CREDENTIALS = "demo-user:demo-user"
+    CREDENTIALS = "demo:demo"
     ftpd = CurlDownload(self.PROTOCOL, SERVER, DIRECTORY)
     ftpd.set_options(dict(ssl_verifyhost="False", ssl_verifypeer="False"))
     ftpd.set_credentials(CREDENTIALS)
@@ -686,7 +992,7 @@ class TestBiomajFTPSDownload(unittest.TestCase):
     # This server is misconfigured hence we disable all SSL verification
     SERVER = "demo.wftpserver.com"
     DIRECTORY = "/download/"
-    CREDENTIALS = "demo-user:demo-user"
+    CREDENTIALS = "demo:demo"
     ftpd = CurlDownload(self.PROTOCOL, SERVER, DIRECTORY)
     ftpd.set_options(dict(ssl_verifyhost="False", ssl_verifypeer="False"))
     ftpd.set_credentials(CREDENTIALS)
@@ -696,12 +1002,12 @@ class TestBiomajFTPSDownload(unittest.TestCase):
     ftpd.close()
     self.assertTrue(len(ftpd.files_to_download) == 1)
 
-  def test_download_ssl_certficate(self):
+  def test_download_ssl_certificate(self):
     # This server is misconfigured but we use its certificate
     # The hostname is wrong so we disable host verification
     SERVER = "demo.wftpserver.com"
     DIRECTORY = "/download/"
-    CREDENTIALS = "demo-user:demo-user"
+    CREDENTIALS = "demo:demo"
     ftpd = CurlDownload(self.PROTOCOL, SERVER, DIRECTORY)
     curdir = os.path.dirname(os.path.realpath(__file__))
     cert_file = os.path.join(curdir, "caert.demo.wftpserver.com.pem")
@@ -735,6 +1041,16 @@ class TestBiomajRSYNCDownload(unittest.TestCase):
         (files_list, dir_list) = rsyncd.list()
         self.assertTrue(len(files_list) != 0)
 
+    def test_rsync_list_error(self):
+        # Access a non-existent directory
+        rsyncd = RSYNCDownload("/tmp/foo/", "")
+        # Check that we raise an exception and log a message
+        with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+            with self.assertRaises(Exception):
+                (file_list, dir_list) = rsyncd.list()
+            # Test log message format (we assume that there is only 1 message)
+            self.assertRegex(cm.output[0], "Error while listing")
+
     def test_rsync_match(self):
         rsyncd = RSYNCDownload(self.examples, "")
         (files_list, dir_list) = rsyncd.list()
@@ -772,6 +1088,51 @@ class TestBiomajRSYNCDownload(unittest.TestCase):
         rsyncd.download(self.utils.data_dir)
         self.assertTrue(len(rsyncd.files_to_download) == 3)
 
+    def test_rsync_download_skip_check_uncompress(self):
+        """
+        Download the fake archive file with RSYNC but skip check.
+        """
+        rsyncd = RSYNCDownload(self.utils.test_dir + '/', "")
+        rsyncd.set_options(dict(skip_check_uncompress=True))
+        (file_list, dir_list) = rsyncd.list()
+        rsyncd.match([r'invalid.gz'], file_list, dir_list, prefix='')
+        rsyncd.download(self.utils.data_dir)
+        self.assertTrue(len(rsyncd.files_to_download) == 1)
+
+    def test_rsync_download_retry(self):
+        """
+        Try to download fake files to test retry.
+        """
+        n_attempts = 5
+        rsyncd = RSYNCDownload(self.utils.test_dir + '/', "")
+        rsyncd.set_options(dict(skip_check_uncompress=True))
+        # Download a fake file
+        rsyncd.set_files_to_download([
+              {'name': 'TOTO.zip', 'year': '2016', 'month': '02', 'day': '19',
+               'size': 1, 'save_as': 'TOTO1KB'}
+        ])
+        rsyncd.set_options(dict(stop_condition=tenacity.stop.stop_after_attempt(n_attempts),
+                                wait_condition=tenacity.wait.wait_none()))
+        self.assertRaisesRegex(
+            Exception, "^RSYNCDownload:Download:Error:",
+            rsyncd.download, self.utils.data_dir,
+        )
+        logging.debug(rsyncd.retryer.statistics)
+        self.assertTrue(len(rsyncd.files_to_download) == 1)
+        self.assertTrue(rsyncd.retryer.statistics["attempt_number"] == n_attempts)
+        # Try to download another file to ensure that it retryies
+        rsyncd.set_files_to_download([
+              {'name': 'TITI.zip', 'year': '2016', 'month': '02', 'day': '19',
+               'size': 1, 'save_as': 'TOTO1KB'}
+        ])
+        self.assertRaisesRegex(
+            Exception, "^RSYNCDownload:Download:Error:",
+            rsyncd.download, self.utils.data_dir,
+        )
+        self.assertTrue(len(rsyncd.files_to_download) == 1)
+        self.assertTrue(rsyncd.retryer.statistics["attempt_number"] == n_attempts)
+        rsyncd.close()
+
 
 class iRodsResult(object):
 
@@ -844,6 +1205,7 @@ class MockiRODSSession(object):
         my_test_file = open("tests/test.fasta.gz", "r+")
         return(my_test_file)
 
+
 @attr('irods')
 @attr('roscoZone')
 @attr('network')
@@ -872,18 +1234,107 @@ class TestBiomajIRODSDownload(unittest.TestCase):
         (files_list, dir_list) = irodsd.list()
         self.assertTrue(len(files_list) != 0)
 
-    @attr('local_irods')
+
+@attr('local_irods')
+@attr('network')
+class TestBiomajLocalIRODSDownload(unittest.TestCase):
+    """
+    Test with a local iRODS server.
+    """
+
+    def setUp(self):
+        self.utils = UtilsForLocalIRODSTest()
+        self.curdir = os.path.dirname(os.path.realpath(__file__))
+        self.examples = os.path.join(self.curdir,'bank') + '/'
+        BiomajConfig.load_config(self.utils.global_properties, allow_user_config=False)
+
+    def tearDown(self):
+        self.utils.clean()
+
     def test_irods_download(self):
-        # To run this test, you need an iRODS server on localhost (default
-        # port, user 'rods', password 'rods'), and populate a zone
-        # /tempZone/home/rods with a file that matches r'^test.*\.gz$' (for
-        # instance, by copying tests/bank/test/test.fasta.gz).
-        irodsd = IRODSDownload("localhost", "/tempZone/home/rods")
+        irodsd = IRODSDownload(self.utils.SERVER, self.utils.COLLECTION)
         irodsd.set_param(dict(
-            user='rods',
-            password='rods',
+            user=self.utils.USER,
+            password=self.utils.PASSWORD,
         ))
         (file_list, dir_list) = irodsd.list()
         irodsd.match([r'^test.*\.gz$'], file_list, dir_list, prefix='')
         irodsd.download(self.utils.data_dir)
         self.assertTrue(len(irodsd.files_to_download) == 1)
+
+    def test_irods_download_skip_check_uncompress(self):
+        """
+        Download the fake archive file with iRODS but skip check.
+        """
+        irodsd = IRODSDownload(self.utils.SERVER, self.utils.COLLECTION)
+        irodsd.set_options(dict(skip_check_uncompress=True))
+        irodsd.set_param(dict(
+            user=self.utils.USER,
+            password=self.utils.PASSWORD,
+        ))
+        (file_list, dir_list) = irodsd.list()
+        irodsd.match([r'invalid.gz$'], file_list, dir_list, prefix='')
+        irodsd.download(self.utils.data_dir)
+        self.assertTrue(len(irodsd.files_to_download) == 1)
+
+    def test_irods_download_retry(self):
+        """
+        Try to download fake files to test retry.
+        """
+        n_attempts = 5
+        irodsd = IRODSDownload(self.utils.SERVER, self.utils.COLLECTION)
+        irodsd.set_options(dict(skip_check_uncompress=True))
+        irodsd.set_param(dict(
+            user=self.utils.USER,
+            password=self.utils.PASSWORD,
+        ))
+        # Download a fake file
+        irodsd.set_files_to_download([
+              {'name': 'TOTO.zip', 'year': '2016', 'month': '02', 'day': '19',
+               'size': 1, 'save_as': 'TOTO1KB'}
+        ])
+        irodsd.set_options(dict(stop_condition=tenacity.stop.stop_after_attempt(n_attempts),
+                                wait_condition=tenacity.wait.wait_none()))
+        self.assertRaisesRegex(
+            Exception, "^IRODSDownload:Download:Error:",
+            irodsd.download, self.utils.data_dir,
+        )
+        logging.debug(irodsd.retryer.statistics)
+        self.assertTrue(len(irodsd.files_to_download) == 1)
+        self.assertTrue(irodsd.retryer.statistics["attempt_number"] == n_attempts)
+        # Try to download another file to ensure that it retryies
+        irodsd.set_files_to_download([
+              {'name': 'TITI.zip', 'year': '2016', 'month': '02', 'day': '19',
+               'size': 1, 'save_as': 'TOTO1KB'}
+        ])
+        self.assertRaisesRegex(
+            Exception, "^IRODSDownload:Download:Error:",
+            irodsd.download, self.utils.data_dir,
+        )
+        self.assertTrue(len(irodsd.files_to_download) == 1)
+        self.assertTrue(irodsd.retryer.statistics["attempt_number"] == n_attempts)
+        irodsd.close()
+
+    def test_irods_list_error(self):
+        # Non-existing collection
+        irodsd = IRODSDownload(self.utils.SERVER, "fake_collection")
+        irodsd.set_param(dict(
+            user=self.utils.USER,
+            password=self.utils.PASSWORD,
+        ))
+        with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+            with self.assertRaises(Exception):
+                (file_list, dir_list) = irodsd.list()
+            # Test log message format (we assume that there is only 1 message)
+            self.assertRegex(cm.output[0], "Error while listing")
+        # Test with wrong password
+        irodsd = IRODSDownload(self.utils.SERVER, self.utils.COLLECTION)
+        irodsd.set_param(dict(
+            user=self.utils.USER,
+            password="badpassword",
+        ))
+        with self.assertLogs(logger="biomaj", level="ERROR") as cm:
+            with self.assertRaises(Exception):
+                (file_list, dir_list) = irodsd.list()
+            # Test log message format (we assume that there is only 1 message)
+            self.assertRegex(cm.output[0], "Error while listing")
